/*
 * Copyright 2011 Steven Gribble
 *
 *  This file is part of the UW CSE 333 course project sequence
 *  (333proj).
 *
 *  333proj is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  333proj is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <assert.h>

#include "hashtable.h"
#include "hashtable_priv.h"

// A private utility function to grow the hashtable (increase
// the number of buckets) if its load factor has become too high.
static void ResizeHashtable(HashTableRecordPtr ht);

// a free function that does nothing
static void NullFree(void *freeme) { }

HashTable AllocateHashTable(uint32_t num_buckets) {
  HashTableRecordPtr ht;
  uint32_t            i;

  // defensive programming
  if (num_buckets == 0) {
    return NULL;
  }

  // allocate the hash table record
  ht = (HashTableRecordPtr) malloc(sizeof(HashTableRecord));
  if (ht == NULL) {
    return NULL;
  }

  // initialize the record
  ht->num_buckets = num_buckets;
  ht->num_elements = 0;
  ht->buckets =
    (LinkedList *) malloc(num_buckets * sizeof(LinkedList));
  if (ht->buckets == NULL) {
    // make sure we don't leak!
    free(ht);
    return NULL;
  }
  for (i = 0; i < num_buckets; i++) {
    ht->buckets[i] = AllocateLinkedList();
    if (ht->buckets[i] == NULL) {
      // allocating one of our bucket chain lists failed,
      // so we need to free everything we allocated so far
      // before returning NULL to indicate failure.  Since
      // we know the chains are empty, we'll pass in a
      // free function pointer that does nothing; it should
      // never be called.
      uint32_t j;
      for (j = 0; j < i; j++) {
        FreeLinkedList(ht->buckets[j], NullFree);
      }
      free(ht);
      return NULL;
    }
  }

  return (HashTable) ht;
}

void FreeHashTable(HashTable table,
                   ValueFreeFnPtr value_free_function) {
  HashTableRecordPtr ht = (HashTableRecordPtr) table;
  uint32_t i;

  assert(ht != NULL);  // be defensive

  // loop through and free the chains on each bucket
  for (i = 0; i < ht->num_buckets; i++) {
    LinkedList  bl = ht->buckets[i];
    HTKeyValue *nextKV;

    // pop elements off the the chain list, then free the list
    while (NumElementsInLinkedList(bl) > 0) {
      assert(PopLinkedList(bl, (void **) &nextKV));
      value_free_function(nextKV->value);
      free(nextKV);
    }
    // the chain list is empty, so we can pass in the
    // null free function to FreeLinkedList.
    FreeLinkedList(bl, NullFree);
  }

  // free the buckets within the table record,
  // then free the table record itself.
  free(ht->buckets);
  free(ht);
}

uint64_t NumElementsInHashTable(HashTable table) {
  HashTableRecordPtr ht = (HashTableRecordPtr) table;

  assert(ht != NULL);
  return ht->num_elements;
}

int InsertHashTable(HashTable table,
                    HTKeyValue newkeyvalue,
                    HTKeyValue *oldkeyvalue) {
  HashTableRecordPtr ht = (HashTableRecordPtr) table;
  uint32_t insertbucket;
  LinkedList insertchain;

  assert(ht != NULL);
  ResizeHashtable(ht);

  // calculate which bucket we're inserting into,
  // grab its linked list chain
  insertbucket = HashKeyToBucketNum(ht, newkeyvalue.key);
  insertchain = ht->buckets[insertbucket];

  // STEP 1 -- finish the implementation of InsertHashTable.


  return 0;  // you may need to change this return value.
}

int LookupHashTable(HashTable table,
                    uint64_t key,
                    HTKeyValue *keyvalue) {
  HashTableRecordPtr ht = (HashTableRecordPtr) table;

  assert(ht != NULL);
  // STEP 2 -- implement LookupHashTable.


  return 0;  // you may need to change this return value.
}

int RemoveFromHashTable(HashTable table,
                        uint64_t key,
                        HTKeyValue *keyvalue) {
  HashTableRecordPtr ht = (HashTableRecordPtr) table;

  assert(ht != NULL);
  // STEP 3 -- implement RemoveFromHashTable.


  return 0;  // you may need to change this return value.
}

HTIter HashTableMakeIterator(HashTable table) {
  HashTableRecordPtr ht = (HashTableRecordPtr) table;
  HTIterRecord *iter;
  uint32_t      i;

  assert(ht != NULL);  // be defensive

  // malloc the iterator
  iter = (HTIterRecord *) malloc(sizeof(HTIterRecord));
  if (iter == NULL) {
    return NULL;
  }

  // if the hash table is empty, the iterator is immediately invalid,
  // since it can't point to anything.
  if (ht->num_elements == 0) {
    iter->is_valid = 0;
    iter->ht = ht;
    iter->bucket_it = NULL;
    return iter;
  }

  // initialize the iterator.  there is at least one element in the
  // table, so find the first element and point the iterator at it.
  iter->is_valid = 1;
  iter->ht = ht;
  for (i = 0; i < ht->num_buckets; i++) {
    if (NumElementsInLinkedList(ht->buckets[i]) > 0) {
      iter->bucket_num = i;
      break;
    }
  }
  assert(i < ht->num_buckets);  // make sure we found it.
  iter->bucket_it =
    LinkedListMakeIterator(ht->buckets[iter->bucket_num], 0);
  if (iter->bucket_it == NULL) {
    // out of memory!
    free(iter);
    return NULL;
  }
  return iter;
}

void HTIteratorFree(HTIter iter) {
  HTIterRecord *it = (HTIterRecord *) iter;

  assert(it != NULL);
  if (it->bucket_it != NULL) {
    LLIteratorFree(it->bucket_it);
    it->bucket_it = NULL;
  }
  it->is_valid = 0;
  free(it);
}

int HTIteratorNext(HTIter iter) {
  HTIterRecord *it = (HTIterRecord *) iter;

  assert(it != NULL);
  // STEP 4 -- implement HTIteratorNext.


  return 0;  // you might need to change this return value.
}

int HTIteratorPastEnd(HTIter iter) {
  HTIterRecord *it = (HTIterRecord *) iter;

  assert(it != NULL);
  // STEP 5 -- implement HTIteratorPastEnd.


  return 0;  // you might need to change this return value.
}

int HTIteratorGet(HTIter iter, HTKeyValue *keyvalue) {
  HTIterRecord *it = (HTIterRecord *) iter;

  assert(it != NULL);
  // STEP 6 -- implement HTIteratorGet.


  return 0;  // you might need to change this return value.
}

int HTIteratorDelete(HTIter iter, HTKeyValue *keyvalue) {
  HTIterRecord *it = (HTIterRecord *) iter;
  HTKeyValue kv;
  int res, retval;

  assert(it != NULL);

  // Try to get what the iterator is pointing to.
  res = HTIteratorGet(iter, &kv);
  if (res == 0)
    return 0;

  // Advance the iterator.
  res = HTIteratorNext(iter);
  if (res == 0) {
    retval = 2;
  } else {
    retval = 1;
  }
  res = RemoveFromHashTable(it->ht, kv.key, keyvalue);
  assert(res == 1);
  assert(kv.key == keyvalue->key);
  assert(kv.value == keyvalue->value);

  return retval;
}

uint64_t FNVHash64(unsigned char *buffer, unsigned int len) {
  // This code is adapted from code by Landon Curt Noll
  // and Bonelli Nicola:
  //
  // http://code.google.com/p/nicola-bonelli-repo/
  static const uint64_t FNV1_64_INIT = 0xcbf29ce484222325ULL;
  static const uint64_t FNV_64_PRIME = 0x100000001b3ULL;
  unsigned char *bp = (unsigned char *) buffer;
  unsigned char *be = bp + len;
  uint64_t hval = FNV1_64_INIT;

  /*
   * FNV-1a hash each octet of the buffer
   */
  while (bp < be) {
    /* xor the bottom with the current octet */
    hval ^= (uint64_t) * bp++;
    /* multiply by the 64 bit FNV magic prime mod 2^64 */
    hval *= FNV_64_PRIME;
  }
  /* return our new hash value */
  return hval;
}

uint64_t FNVHashInt64(uint64_t hashme) {
  unsigned char buf[8];
  int i;

  for (i = 0; i < 8; i++) {
    buf[i] = (unsigned char) (hashme & 0x00000000000000FFULL);
    hashme >>= 8;
  }
  return FNVHash64(buf, 8);
}

uint32_t HashKeyToBucketNum(HashTableRecordPtr ht, uint64_t key) {
  return (uint32_t) (key % ((uint64_t) ht->num_buckets) );
}

static void ResizeHashtable(HashTableRecordPtr ht) {
  // Resize if the load factor is > 3.
  if (ht->num_elements < 3 * ht->num_buckets)
    return;

  // This is the resize case.  Allocate a new hashtable,
  // iterate over the old hashtable, do the surgery on
  // the old hashtable record and free up the new hashtable
  // record.
  HashTableRecordPtr newht = (HashTableRecordPtr)
    AllocateHashTable(ht->num_buckets * 9);

  // Give up if out of memory.
  if (newht == NULL)
    return;

  // Loop through the old ht with an iterator,
  // inserting into the new HT.
  HTIter it = HashTableMakeIterator(ht);
  if (it == NULL) {
    // Give up if out of memory.
    FreeHashTable(newht, &NullFree);
    return;
  }

  while (!HTIteratorPastEnd(it)) {
    HTKeyValue item, dummy;

    assert(HTIteratorGet(it, &item) == 1);
    if (InsertHashTable(newht, item, &dummy) != 1) {
      // failure, free up everything, return.
      HTIteratorFree(it);
      FreeHashTable(newht, &NullFree);
      return;
    }
    HTIteratorNext(it);
  }

  // Worked!  Free the iterator.
  HTIteratorFree(it);

  // Sneaky: swap the structures, then free the new table,
  // and we're done.
  {
    HashTableRecord tmp;

    tmp = *ht;
    *ht = *newht;
    *newht = tmp;
    FreeHashTable(newht, &NullFree);
  }

  return;
}
