/*
 * Copyright 2011 Steven Gribble
 *
 *  This file is part of the UW CSE 333 course project sequence
 *  (333proj).
 *
 *  333proj is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  333proj is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "ll.h"
#include "ll_priv.h"  // we're white-box testing; clients shouldn't include

#include "CUnit/Basic.h"

// our CUnit suite initialization/cleanup functions
void LLSAddPoints(int points);
int LLSInit(void);
int LLSCleanup(void);

// our CUnit test functions
void LLSTestAllocFree(void);
void LLTestPushAppendPopSlice(void);
void LLTestSort(void);
void LLTestIteratorBasic(void);
void LLTestIteratorDeleteInsert(void);

// our main function
int main(int argc, char **argv) {
  CU_pSuite pSuite = NULL;

  // initialize the test registry
  if (CU_initialize_registry() != CUE_SUCCESS) {
    return CU_get_error();
  }

  // add a suite to the registry
  pSuite = CU_add_suite("linked list suite", LLSInit, LLSCleanup);

  if (pSuite == NULL) {
    CU_cleanup_registry();
    return CU_get_error();
  }

  // add the tests to the suite
  if ((CU_ADD_TEST(pSuite, LLSTestAllocFree) == NULL) ||
      (CU_ADD_TEST(pSuite, LLTestPushAppendPopSlice) == NULL) ||
      (CU_ADD_TEST(pSuite, LLTestSort) == NULL) ||
      (CU_ADD_TEST(pSuite, LLTestIteratorBasic) == NULL) ||
      (CU_ADD_TEST(pSuite, LLTestIteratorDeleteInsert) == NULL)) {
    CU_cleanup_registry();
    return CU_get_error();
  }

  // run all of the tests using the CUnit basic interface
  CU_basic_set_mode(CU_BRM_VERBOSE);
  CU_basic_run_tests();
  CU_cleanup_registry();
  return CU_get_error();
}


/*** OUR TEST FUNCTIONS AND SUITE INITIALIZATION GO BELOW HERE ***/

// our payload structure
typedef struct payload_st {
  int magic_num;
  int payload_num;
} Payload;

// number of payload frees
static int num_payload_frees;

// our payload free test function
void TestPayloadFree(void *payload) {
  num_payload_frees++;
  CU_ASSERT_EQUAL_FATAL(((Payload *) payload)->magic_num, 0xDEADBEEF);
  free(payload);
}

// our payload comparator function
int TestPayloadComparator(void *p1, void *p2) {
  int i1 = ((Payload *) p1)->payload_num;
  int i2 = ((Payload *) p2)->payload_num;

  if (i1 > i2)
    return 1;
  if (i1 < i2)
    return -1;
  return 0;
}


static const int MAXGRADE = 140;
static int yourgrade = 0;

void LLSAddPoints(int points) {
  yourgrade += points;
  printf("(%d/%d)", yourgrade, MAXGRADE);
}

void LLSFinalGrade() {
  printf("\n\nYour unit test grade is %d/%d [%.02f%%]\n",
         yourgrade, MAXGRADE, (100.0*yourgrade)/MAXGRADE);
}


int LLSInit(void) {
  yourgrade = 0;
  printf("There are %d unit-testing points available.\n",
         MAXGRADE);
  return 0;  // return 0 on success, non-zero otherwise
}

int LLSCleanup(void) {
  LLSFinalGrade();
  return 0;  // return 0 on success, non-zero otherwise
}

void LLSTestAllocFree(void) {
  // simple create / delete test
  LinkedList list = AllocateLinkedList();
  LinkedListRecordPtr ll = (LinkedListRecordPtr) list;
  CU_ASSERT_EQUAL_FATAL(ll->num_elements, 0);
  CU_ASSERT_PTR_NULL_FATAL(ll->head);
  CU_ASSERT_PTR_NULL_FATAL(ll->tail);
  CU_ASSERT_EQUAL_FATAL(NumElementsInLinkedList(list), 0);
  LLSAddPoints(10);
  FreeLinkedList(list, &TestPayloadFree);
  LLSAddPoints(10);
}

void LLTestPushAppendPopSlice(void) {
  // test push, append, pop, slice
  LinkedList list = AllocateLinkedList();
  LinkedListRecordPtr ll = (LinkedListRecordPtr) list;
  Payload *p1, *p2, *p3, *p4, *ret;

  // set up some payloads
  p1 = (Payload *) malloc(sizeof(Payload));
  p2 = (Payload *) malloc(sizeof(Payload));
  p3 = (Payload *) malloc(sizeof(Payload));
  p4 = (Payload *) malloc(sizeof(Payload));
  p1->magic_num = p2->magic_num = 0xDEADBEEF;
  p3->magic_num = p4->magic_num =  0xDEADBEEF;
  p1->payload_num = 10;
  p2->payload_num = 5;
  p3->payload_num = 15;
  p4->payload_num = 8;

  // push the first payload on (tests push onto empty list)
  CU_ASSERT_EQUAL_FATAL(PushLinkedList(list, p1), 1);
  CU_ASSERT_EQUAL_FATAL(ll->num_elements, 1);
  CU_ASSERT_PTR_NOT_NULL_FATAL(ll->head);
  CU_ASSERT_PTR_NOT_NULL_FATAL(ll->tail);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head, ll->tail);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->payload, p1);
  CU_ASSERT_PTR_NULL_FATAL(ll->head->next);
  CU_ASSERT_PTR_NULL_FATAL(ll->head->prev);
  LLSAddPoints(5);

  // push the second payload on
  CU_ASSERT_EQUAL_FATAL(PushLinkedList(list, p2), 1);
  CU_ASSERT_EQUAL_FATAL(ll->num_elements, 2);
  CU_ASSERT_PTR_NOT_NULL_FATAL(ll->head);
  CU_ASSERT_PTR_NOT_NULL_FATAL(ll->tail);
  CU_ASSERT_PTR_NOT_EQUAL_FATAL(ll->head, ll->tail);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->payload, p2);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->tail->payload, p1);
  CU_ASSERT_PTR_NULL_FATAL(ll->head->prev);
  CU_ASSERT_PTR_NULL_FATAL(ll->tail->next);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->next, ll->tail);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->tail->prev, ll->head);
  LLSAddPoints(5);
  // push the third payload on
  CU_ASSERT_EQUAL_FATAL(PushLinkedList(list, p3), 1);

  // use append to add the fourth payload to the end
  CU_ASSERT_EQUAL_FATAL(AppendLinkedList(list, p4), 1);
  CU_ASSERT_EQUAL_FATAL(ll->num_elements, 4);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->payload, p3);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->tail->payload, p4);
  CU_ASSERT_PTR_NULL_FATAL(ll->head->prev);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->next->payload, p2);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->next->prev->payload, p3);
  CU_ASSERT_PTR_NULL_FATAL(ll->tail->next);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->tail->prev->payload, p1);
  LLSAddPoints(5);

  // pop from the head
  CU_ASSERT_EQUAL_FATAL(PopLinkedList(list, (void **) &ret), 1);
  CU_ASSERT_PTR_EQUAL_FATAL(ret, p3);
  LLSAddPoints(5);

  // slice from the tail
  CU_ASSERT_EQUAL_FATAL(SliceLinkedList(list, (void **) &ret), 1);
  CU_ASSERT_PTR_EQUAL_FATAL(ret, p4);
  LLSAddPoints(5);

  // pop remaining elements off
  CU_ASSERT_EQUAL_FATAL(PopLinkedList(list, (void **) &ret), 1);
  CU_ASSERT_PTR_EQUAL_FATAL(ret, p2);
  CU_ASSERT_EQUAL_FATAL(PopLinkedList(list, (void **) &ret), 1);
  CU_ASSERT_PTR_EQUAL_FATAL(ret, p1);
  CU_ASSERT_EQUAL_FATAL(ll->num_elements, 0);
  LLSAddPoints(5);

  // try to pop off an empty list, slice off an empty list
  CU_ASSERT_EQUAL_FATAL(PopLinkedList(list, (void **) &ret), 0);
  CU_ASSERT_EQUAL_FATAL(ll->num_elements,  0);
  CU_ASSERT_EQUAL_FATAL(SliceLinkedList(list, (void **) &ret), 0);
  CU_ASSERT_EQUAL_FATAL(ll->num_elements,  0);
  LLSAddPoints(5);

  // append element back
  CU_ASSERT_EQUAL_FATAL(AppendLinkedList(list, p4), 1);
  CU_ASSERT_EQUAL_FATAL(ll->num_elements, 1);
  CU_ASSERT_PTR_NOT_NULL_FATAL(ll->head);
  CU_ASSERT_PTR_NOT_NULL_FATAL(ll->tail);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head, ll->tail);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->payload, p4);
  CU_ASSERT_PTR_NULL_FATAL(ll->head->next);
  CU_ASSERT_PTR_NULL_FATAL(ll->head->prev);
  LLSAddPoints(5);

  // try slicing last element corner case
  CU_ASSERT_EQUAL_FATAL(SliceLinkedList(list, (void **) &ret), 1);
  CU_ASSERT_PTR_EQUAL_FATAL(ret, p4);
  CU_ASSERT_EQUAL_FATAL(ll->num_elements, 0);
  CU_ASSERT_PTR_NULL_FATAL(ll->head);
  CU_ASSERT_PTR_NULL_FATAL(ll->tail);
  LLSAddPoints(5);

  // append element back, do FreeLinkedList
  CU_ASSERT_EQUAL_FATAL(AppendLinkedList(list, p4), 1);

  num_payload_frees = 0;
  FreeLinkedList(list, &TestPayloadFree);
  CU_ASSERT_EQUAL_FATAL(num_payload_frees, 1);
  free(p1);
  free(p2);
  free(p3);
  LLSAddPoints(5);
}

void LLTestSort(void) {
  // test push, append, pop, slice
  LinkedList list = AllocateLinkedList();
  LinkedListRecordPtr ll = (LinkedListRecordPtr) list;
  Payload *p1, *p2, *p3, *p4;

  // set up some payloads
  p1 = (Payload *) malloc(sizeof(Payload));
  p2 = (Payload *) malloc(sizeof(Payload));
  p3 = (Payload *) malloc(sizeof(Payload));
  p4 = (Payload *) malloc(sizeof(Payload));
  p1->magic_num = p2->magic_num = 0xDEADBEEF;
  p3->magic_num = p4->magic_num =  0xDEADBEEF;
  p1->payload_num = 10;
  p2->payload_num = 5;
  p3->payload_num = 15;
  p4->payload_num = 8;

  // push our payloads on
  CU_ASSERT_EQUAL_FATAL(AppendLinkedList(list, p1), 1);
  CU_ASSERT_EQUAL_FATAL(AppendLinkedList(list, p2), 1);
  CU_ASSERT_EQUAL_FATAL(AppendLinkedList(list, p3), 1);
  CU_ASSERT_EQUAL_FATAL(AppendLinkedList(list, p4), 1);

  // sort ascending
  SortLinkedList(list, 1, TestPayloadComparator);
  CU_ASSERT_EQUAL_FATAL(ll->head->payload, p2);
  CU_ASSERT_EQUAL_FATAL(ll->head->next->payload, p4);
  CU_ASSERT_EQUAL_FATAL(ll->head->next->next->payload, p1);
  CU_ASSERT_EQUAL_FATAL(ll->head->next->next->next->payload, p3);
  CU_ASSERT_PTR_NULL_FATAL(ll->head->next->next->next->next);
  CU_ASSERT_EQUAL_FATAL(ll->tail->payload, p3);
  CU_ASSERT_EQUAL_FATAL(ll->tail->prev->payload, p1);
  CU_ASSERT_EQUAL_FATAL(ll->tail->prev->prev->payload, p4);
  CU_ASSERT_EQUAL_FATAL(ll->tail->prev->prev->prev->payload, p2);
  CU_ASSERT_PTR_NULL_FATAL(ll->tail->prev->prev->prev->prev);

  // sort descending
  SortLinkedList(list, 0, TestPayloadComparator);
  CU_ASSERT_EQUAL_FATAL(ll->head->payload, p3);
  CU_ASSERT_EQUAL_FATAL(ll->head->next->payload, p1);
  CU_ASSERT_EQUAL_FATAL(ll->head->next->next->payload, p4);
  CU_ASSERT_EQUAL_FATAL(ll->head->next->next->next->payload, p2);
  CU_ASSERT_PTR_NULL_FATAL(ll->head->next->next->next->next);
  CU_ASSERT_EQUAL_FATAL(ll->tail->payload, p2);
  CU_ASSERT_EQUAL_FATAL(ll->tail->prev->payload, p4);
  CU_ASSERT_EQUAL_FATAL(ll->tail->prev->prev->payload, p1);
  CU_ASSERT_EQUAL_FATAL(ll->tail->prev->prev->prev->payload, p3);
  CU_ASSERT_PTR_NULL_FATAL(ll->tail->prev->prev->prev->prev);

  // free the list
  num_payload_frees = 0;
  FreeLinkedList(list, &TestPayloadFree);
  CU_ASSERT_EQUAL_FATAL(num_payload_frees, 4);
}

void LLTestIteratorBasic(void) {
  // test push, append, pop, slice
  LinkedList list = AllocateLinkedList();
  LinkedListRecordPtr ll = (LinkedListRecordPtr) list;
  LLIterRecordPtr iter;
  Payload *p1, *p2, *p3, *p4, *ret;

  // set up some payloads
  p1 = (Payload *) malloc(sizeof(Payload));
  p2 = (Payload *) malloc(sizeof(Payload));
  p3 = (Payload *) malloc(sizeof(Payload));
  p4 = (Payload *) malloc(sizeof(Payload));
  p1->magic_num = p2->magic_num = 0xDEADBEEF;
  p3->magic_num = p4->magic_num =  0xDEADBEEF;
  p1->payload_num = 10;
  p2->payload_num = 5;
  p3->payload_num = 15;
  p4->payload_num = 8;

  // push our payloads on
  CU_ASSERT_EQUAL_FATAL(AppendLinkedList(list, p1), 1);
  CU_ASSERT_EQUAL_FATAL(AppendLinkedList(list, p2), 1);
  CU_ASSERT_EQUAL_FATAL(AppendLinkedList(list, p3), 1);
  CU_ASSERT_EQUAL_FATAL(AppendLinkedList(list, p4), 1);
  CU_ASSERT_EQUAL_FATAL(ll->num_elements, 4);

  // make the iterator from the head
  iter = (LLIterRecordPtr) LinkedListMakeIterator(list, 0);
  CU_ASSERT_PTR_NOT_NULL_FATAL(iter);
  CU_ASSERT_PTR_EQUAL_FATAL(iter->list, ll);
  CU_ASSERT_PTR_EQUAL_FATAL(iter->node, ll->head);
  CU_ASSERT_EQUAL_FATAL(LLIteratorAtHead(iter), 1);
  CU_ASSERT_EQUAL_FATAL(LLIteratorAtTail(iter), 0);
  LLSAddPoints(10);

  // free the iterator
  LLIteratorFree(iter);
  LLSAddPoints(5);

  // make the iterator from the tail
  iter = (LLIterRecordPtr) LinkedListMakeIterator(list, 1);
  CU_ASSERT_PTR_NOT_NULL_FATAL(iter);
  CU_ASSERT_PTR_EQUAL_FATAL(iter->list, ll);
  CU_ASSERT_PTR_EQUAL_FATAL(iter->node, ll->tail);
  CU_ASSERT_EQUAL_FATAL(LLIteratorAtTail(iter), 1);
  CU_ASSERT_EQUAL_FATAL(LLIteratorAtHead(iter), 0);
  LLSAddPoints(5);

  // try navigating the iterator
  CU_ASSERT_EQUAL_FATAL(LLIteratorNext(iter), 0);   // off end of list
  CU_ASSERT_EQUAL_FATAL(LLIteratorGetPayload(iter, (void **) &ret), 1);
  CU_ASSERT_PTR_EQUAL_FATAL(ret, p4);
  CU_ASSERT_EQUAL_FATAL(LLIteratorPrev(iter), 1);
  CU_ASSERT_EQUAL_FATAL(LLIteratorGetPayload(iter, (void **) &ret), 1);
  CU_ASSERT_PTR_EQUAL_FATAL(ret, p3);
  CU_ASSERT_EQUAL_FATAL(LLIteratorNext(iter), 1);
  CU_ASSERT_EQUAL_FATAL(LLIteratorGetPayload(iter, (void **) &ret), 1);
  CU_ASSERT_PTR_EQUAL_FATAL(ret, p4);
  CU_ASSERT_EQUAL_FATAL(LLIteratorNext(iter), 0);   // off end of list
  CU_ASSERT_EQUAL_FATAL(LLIteratorGetPayload(iter, (void **) &ret), 1);
  CU_ASSERT_PTR_EQUAL_FATAL(ret, p4);
  CU_ASSERT_EQUAL_FATAL(LLIteratorPrev(iter), 1);
  CU_ASSERT_EQUAL_FATAL(LLIteratorGetPayload(iter, (void **) &ret), 1);
  CU_ASSERT_PTR_EQUAL_FATAL(ret, p3);
  CU_ASSERT_EQUAL_FATAL(LLIteratorPrev(iter), 1);
  CU_ASSERT_EQUAL_FATAL(LLIteratorGetPayload(iter, (void **) &ret), 1);
  CU_ASSERT_PTR_EQUAL_FATAL(ret, p2);
  CU_ASSERT_EQUAL_FATAL(LLIteratorPrev(iter), 1);
  CU_ASSERT_EQUAL_FATAL(LLIteratorGetPayload(iter, (void **) &ret), 1);
  CU_ASSERT_PTR_EQUAL_FATAL(ret, p1);
  CU_ASSERT_EQUAL_FATAL(LLIteratorPrev(iter), 0);   // off head of list
  CU_ASSERT_EQUAL_FATAL(LLIteratorGetPayload(iter, (void **) &ret), 1);
  CU_ASSERT_PTR_EQUAL_FATAL(ret, p1);

  // free the iterator
  LLIteratorFree(iter);
  LLSAddPoints(10);

  // free the list
  num_payload_frees = 0;
  FreeLinkedList(list, &TestPayloadFree);
  CU_ASSERT_EQUAL_FATAL(num_payload_frees, 4);
}

void LLTestIteratorDeleteInsert(void) {
  // test push, append, pop, slice
  LinkedList list = AllocateLinkedList();
  LinkedListRecordPtr ll = (LinkedListRecordPtr) list;
  LLIterRecordPtr iter;
  Payload *p1, *p2, *p3, *p4;

  // set up some payloads
  p1 = (Payload *) malloc(sizeof(Payload));
  p2 = (Payload *) malloc(sizeof(Payload));
  p3 = (Payload *) malloc(sizeof(Payload));
  p4 = (Payload *) malloc(sizeof(Payload));
  p1->magic_num = p2->magic_num = 0xDEADBEEF;
  p3->magic_num = p4->magic_num =  0xDEADBEEF;
  p1->payload_num = 10;
  p2->payload_num = 5;
  p3->payload_num = 15;
  p4->payload_num = 8;

  // make the iterator from the head
  iter = (LLIterRecordPtr) LinkedListMakeIterator(list, 0);

  // use LLIteratorInsertBefore to add elements
  CU_ASSERT_EQUAL_FATAL(LLIteratorInsertBefore(iter, p4), 1);
  CU_ASSERT_EQUAL_FATAL(ll->num_elements, 1);
  CU_ASSERT_PTR_NOT_NULL_FATAL(ll->head);
  CU_ASSERT_PTR_NOT_NULL_FATAL(ll->tail);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head, ll->tail);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->payload, p4);
  CU_ASSERT_PTR_NULL_FATAL(ll->head->next);
  CU_ASSERT_PTR_NULL_FATAL(ll->head->prev);
  CU_ASSERT_PTR_EQUAL_FATAL(iter->node, ll->head);
  LLSAddPoints(10);

  CU_ASSERT_EQUAL_FATAL(LLIteratorInsertBefore(iter, p1), 1);
  CU_ASSERT_EQUAL_FATAL(ll->num_elements, 2);
  CU_ASSERT_PTR_NOT_NULL_FATAL(ll->head);
  CU_ASSERT_PTR_NOT_NULL_FATAL(ll->tail);
  CU_ASSERT_PTR_NOT_EQUAL_FATAL(ll->head, ll->tail);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->payload, p1);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->tail->payload, p4);
  CU_ASSERT_PTR_NULL_FATAL(ll->head->prev);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->next, ll->tail);
  CU_ASSERT_PTR_EQUAL_FATAL(iter->node, ll->tail);
  LLSAddPoints(5);

  CU_ASSERT_EQUAL_FATAL(LLIteratorInsertBefore(iter, p2), 1);
  CU_ASSERT_EQUAL_FATAL(ll->num_elements, 3);
  CU_ASSERT_PTR_NOT_NULL_FATAL(ll->head);
  CU_ASSERT_PTR_NOT_NULL_FATAL(ll->tail);
  CU_ASSERT_PTR_NOT_EQUAL_FATAL(ll->head, ll->tail);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->payload, p1);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->tail->payload, p4);
  CU_ASSERT_PTR_NULL_FATAL(ll->head->prev);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->next->payload, p2);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->next->prev, ll->head);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->next->next->payload, p4);
  CU_ASSERT_PTR_EQUAL_FATAL(ll->head->next->next->prev, ll->head->next);
  CU_ASSERT_PTR_EQUAL_FATAL(iter->node, ll->tail);
  LLSAddPoints(5);

  CU_ASSERT_EQUAL_FATAL(LLIteratorInsertBefore(iter, p3), 1);
  CU_ASSERT_EQUAL_FATAL(ll->num_elements, 4);

  // use LLIteratorDelete to delete elements
  num_payload_frees = 0;
  CU_ASSERT_EQUAL_FATAL(LLIteratorPrev(iter), 1);
  CU_ASSERT_EQUAL_FATAL(LLIteratorDelete(iter, TestPayloadFree), 1);
  CU_ASSERT_EQUAL_FATAL(num_payload_frees, 1);
  CU_ASSERT_PTR_EQUAL_FATAL(iter->node, ll->tail);
  LLSAddPoints(10);

  CU_ASSERT_EQUAL_FATAL(LLIteratorDelete(iter, TestPayloadFree), 1);
  CU_ASSERT_EQUAL_FATAL(num_payload_frees, 2);
  CU_ASSERT_PTR_EQUAL_FATAL(iter->node, ll->tail);
  LLSAddPoints(5);

  CU_ASSERT_EQUAL_FATAL(LLIteratorDelete(iter, TestPayloadFree), 1);
  CU_ASSERT_EQUAL_FATAL(num_payload_frees, 3);
  CU_ASSERT_PTR_EQUAL_FATAL(iter->node, ll->head);
  CU_ASSERT_PTR_EQUAL_FATAL(iter->node, ll->tail);

  CU_ASSERT_EQUAL_FATAL(LLIteratorDelete(iter, TestPayloadFree), 0);
  CU_ASSERT_EQUAL_FATAL(num_payload_frees, 4);
  CU_ASSERT_PTR_NULL_FATAL(iter->node);
  LLSAddPoints(5);

  // free the iterator
  LLIteratorFree(iter);

  // free the list
  num_payload_frees = 0;
  FreeLinkedList(list, &TestPayloadFree);
  CU_ASSERT_EQUAL_FATAL(num_payload_frees, 0);
}
